function [ f ] = density( x, y, b, T_b, s_b, l_b, T_x, s_x, l_x, r )
%   Estimate the density of parameter distribution at b
%   x = a matrix of parameter values. Rows give different data points;
%   columns give values for different dimensions.
%   y = a vector of realizations. Rows give different data points 0 or 1.
%   b = a matrix of points at which to evaluate the density. Rows give
%   different points; columns give values for different dimensions.
%   T_b = T_N parameter for outside loop
%   s_b = s parameter for outside loop
%   l_b = l parameter for outside loop
%   T_x = T_N parameter for calculation of \hat{f}_X
%   s_x = s parameter for calculation of \hat{f}_X
%   l_x = l parameter for calculation of \hat{f}_X
%   r = truncation parameter for (logN)^(-r)

    xsize = size(x);
    ysize = size(y);
    bsize = size(b);
    
    assert( xsize(1) == ysize(1) ); % same number of data points for (x,y)
    assert( xsize(2) == bsize(2) ); % needed to guarantee dot product xb
    
    
    % dependent constants
    N = xsize(1);
    d = xsize(2);
    nu_d = (d-2)/2.0;
    
    % modify constants to satisfy theoretical constraints
    
    % l is an integer satisfying l > (d-2)/2
    l_b = max( round(l_b), floor((d-2)/2.0)+1);
    l_x = max( round(l_x), floor((d-2)/2.0)+1);
    
    % s  > 0
    s_b = max(s_b, 0);
    s_x = max(s_x, 0);
    
    % r > 0
    r = max(r, 0);
    
    f = zeros(bsize(1), 1); % 1 value for each beta   
    % \sum_{p=0}^{T_N}
    for p = 0:T_b
        multiplier = chi(2*p+1,2*T_b+1,d,s_b,l_b) * h(2*p+1, d);
        multiplier = multiplier/(lambda(2*p+1, d)*c(nu_d, 2*p+1,1));
        multiplier = multiplier / N;
        
        % \sum_{i=1}^{N}
        
        % calculate vectorized numerator
        vec_num_1 = 2*y - 1;
        % vec_num_1 is a column vector for different y_i
        
        vec_num_2 = c(nu_d, 2*p+1, x*(b'));
        % each fixed column of vec_num_2 is for a single beta, iterating
        % through the different x_i
        
        % we need to multiply element by element for every beta
        vec_num_1 = repmat(vec_num_1, 1, bsize(1));
        vec_num = vec_num_1 .* vec_num_2;
        % each column of vec_num is for a different beta.
        % each row is for a different x_i.
        
        % now calculate the denominator
        % first calculate the \hat{f}_X(x_j) for each x_j
        
        inside_total = zeros(1, N);
        for n = 0:T_x
            inside_mult = chi(n,T_x,d,s_x,l_x) * h(n, d);
            inside_mult = inside_mult / (s(d-1) * c(nu_d, n, 1));
            
            % vectorize the sum of C_n^{nu_d} (x_i ' x_j)
            inside_matrix = c(nu_d, n, x * (x'));
            % each column gives value for i = i, ..., N for fixed j
            inside_matrix = sum(inside_matrix);
            % each column is the sum from i = 1...N for fixed j
            inside_matrix = inside_matrix .* inside_mult;
            % we have the full inside term for all j
            
            % sum from n =0...T
            inside_total = inside_total + inside_matrix;
        end
        
        % inside_total is a row vector. each column is for a different j
        
        vec_den = inside_total / N; % denominator absorbs outside 1/N term
        vec_den = max(vec_den, max(0, log(N)^(-r)));
        % each column is for a different j
        % recall that each column of vec_num is for a different beta
        vec_num = vec_num';
        % now each row of vec_num is for a different beta
        
        % the denominator is constant across betas
        vec_den = repmat(vec_den, bsize(1), 1);
        
        vec_inside = vec_num ./ vec_den;
        
        % sum each row to obtain inside term for each beta
        vec_inside = sum(vec_inside, 2);
        
        term_total = multiplier .* vec_inside;
        
        f = f + term_total;
    end
    
    f = max(2 .*f, 0 .* f) ./ s(d-1); % changed from d-1
end

